<?php
namespace AJAXGuestBook\core;

use Carbon\Carbon;
use Carbon\CarbonPeriod;
use DOMDocument;
use DOMXPath;
use SBBCodeParser\Node_Container_Document;

/**
 * @property DOMDocument $xmlDoc
 */
class GBUpgrader {
    private $database; //String: XML file name where all the comments are stored
    private $xmlDoc; //DOM Object: XML object to manipulate guestbook comments data
    private $bbcodeparser;

    function __construct(){
        $this->database = "data.xml";
    }

    public function setDatabaseFile($datafile){
        $this->database = $datafile;
    }

    /* Loading the xml document.*/
    private function loadXMLDoc(){
        if(!file_exists(DATAPATH . DS . $this->database)){
            return false;
        }
        $this->xmlDoc = new DOMDocument();
        $this->xmlDoc->preserveWhiteSpace = false;
        $opened = @$this->xmlDoc->Load(DATAPATH . DS . $this->database);
        return $opened;
    }

    /* Saving the xml document.*/
    private function saveXMLDoc(){
        if(!file_exists(DATAPATH . DS . $this->database)){
            return false;
        }
        $saveddoc = @$this->xmlDoc->save(DATAPATH . DS . $this->database);
        return $saveddoc;
    }

    /*Generate random unique id for message.*/
    private function getRandomUniqueID(){
        do{
            $uniqueid = uniqid("mid");
        } while($this->checkIfUniqueIdExists($uniqueid));
        return $uniqueid;
    }

    private function checkIfUniqueIdExists($uid){
        $messageList = new DOMXpath($this->xmlDoc);
        $uniqid = $messageList->query("/guestbook/message[@uid='" . $uid . "']");
        if($uniqid->length > 0){
            return true;
        } else{
            return false;
        }
    }

    public function upgradedbv4(){
        $loadvalue = $this->loadXMLDoc();
        if($loadvalue == false){
            $returnvalue['error'] = 'Error opening the xml database file.';
            return $returnvalue;
        }
        $guestbook = $this->xmlDoc->documentElement;
        if($guestbook->hasAttribute('dbversion')){
            $dbversion = $guestbook->getAttribute('dbversion');
            if(!((int)$dbversion < DATABASE_VERSION)){
                $returnvalue['info'] = 'Database is already upgraded.';
                return $returnvalue;
            }
        }
        $guestbook->setAttribute('dbversion', DATABASE_VERSION);
        $messageList = $this->xmlDoc->getElementsByTagName('message');
        $messageCount = $messageList->length;
        if($messageCount == 0){
            $returnvalue['error'] = 'Either the database was empty with no entries or the XML database file was invalid.';
            return $returnvalue;
        }
        /* @var $curmessage \DOMElement */
        for($i = $messageCount; $i > 0; $i--){
            $curmessage = $messageList->item($i - 1);
            $approvalstatus = $curmessage->childNodes->item(5);
            if($approvalstatus == null){
                $curmessage->appendChild($this->xmlDoc->createElement('approved', "true"));
            }
            $messageid = $this->getRandomUniqueID();
            $curmessage->setAttribute('uid', $messageid);
            $datetime = $curmessage->getAttribute('datetime');
            $datetime = GBLocale::convertOldDateTimetoTimeStamp($datetime);
            $curmessage->setAttribute('datetime', $datetime);
            $comments = $curmessage->childNodes->item(4)->nodeValue;
            $this->bbcodeparser = new Node_Container_Document(true, false);
            $comments = $this->bbcodeparser->parse($comments)->detect_links()->detect_emails()->get_html();
            $comments = html_entity_decode($comments);
            $comments = CommentPurifier::purify($comments);
            $comments = $this->convertOldSmileys($comments);
            $comnode = $this->xmlDoc->createElement('comments');
            $comnode->appendChild($this->xmlDoc->createCDATASection($comments));
            $curmessage->replaceChild($comnode, $curmessage->childNodes->item(4));
            $curmessage->appendChild($this->xmlDoc->createElement('useragent', ""));
        }
        $saved = $this->saveXMLDoc();
        if($saved == false){
            $returnvalue['error'] = 'Error upgrading database. Please try again.';
            return $returnvalue;
        } else{
            $returnvalue['success'] = 'Database successfully upgraded.';
            return $returnvalue;
        }
    }

    private function convertOldSmileys($message){
        $oldemotes = array('>:(' => ':angry:', '^_^' => ':anime:', '8-)' => ':bigeyes:', ':-D' => ':bigsmile:',
                           ';-D' => ':bigwink:', ':blue:' => ':blue:', 'O.o' => ':boggle:', ':-?' => ':confuse:',
                           'B-)' => ':cool:', '>:-)' => ':evil:', ':-(' => ':frown:', '<3' => ':heart:',
                           ':-/' => ':irritated:', 'X-D' => ':laugh:', '<_<' => ':lookleft:', '>_>' => ':lookright:',
                           ':-l' => ':neutral:', 'O:-)' => ':saint:', ':zzz:' => ':sleepy:', ':-)' => ':smile:',
                           ':-3' => ':smile3:', '>;-)' => ':sneaky:', ':star:' => ':star:', ':-O' => ':surprise:',
                           '^^;' => ':sweatdrop:', '<G>' => ':teeth:', ':P' => ':tongue:', ';-)' => ':wink:',
                           ';-3' => ':wink3:', ':-S' => ':worry:');
        foreach($oldemotes as $oldcode => $newcode){
            $message = str_replace($oldcode, $newcode, $message);
        }
        return $message;
    }

    public function checkDatabaseUpgrade(){
        $loadvalue = $this->loadXMLDoc();
        if($loadvalue == false){
            Session::set("gberror", 'Error opening the xml database file. Please contact and inform the guestbook owner.');
            return false;
        }
        $guestbook = $this->xmlDoc->documentElement;
        $dbversion = $guestbook->getAttribute('dbversion');
        if((int)$dbversion < DATABASE_VERSION){
            return true;
        } else{
            return false;
        }
    }

    public function resetDates(){
        $loadvalue = $this->loadXMLDoc();
        if($loadvalue == false){
            Session::set("gberror", 'Error opening the xml database file. Please contact and inform the guestbook owner.');
            return false;
        }
        $messageList = $this->xmlDoc->getElementsByTagName('message');
        $messageCount = $messageList->length;
        if($messageCount == 0){
            Session::set("gberror", 'Either the database was empty with no entries or the XML database file was invalid.');
            return false;
        }
        $startdate = Carbon::createFromDate(2010, 4, 6);
        $diffDays = $startdate->diffInDays();
        $diffDays = floor($diffDays / $messageCount);
        $enddate = Carbon::now();
        $period = CarbonPeriod::since($startdate)->days($diffDays)->until($enddate);
        $pArry = array_reverse($period->toArray());
        for($i = $messageCount; $i > 0; $i--){
            /* @var $curmessage \DOMElement */
            $curmessage = $messageList->item($i - 1);
            $newTimestamp = $pArry[$i - 1]->timestamp;
            $curmessage->setAttribute('datetime', $newTimestamp);
        }
        $saved = $this->saveXMLDoc();
        if($saved == false){
            Session::set("gberror", 'Error upgrading database. Please contact and inform the guestbook owner.');
            return false;
        } else{
            Session::set("gbsuccess", ' Database successfully upraded.');
            return true;
        }
    }
}